import math


class Size:
    def __init__(self, width, height):
        self.w = width
        self.h = height

    @property
    def width(self):
        return self.w

    @property
    def height(self):
        return self.h

    @width.setter
    def width(self, value):
        self.w = value

    @height.setter
    def height(self, value):
        self.h = value

    @property
    def wh(self):
        return (self.width, self.height)

    @wh.setter
    def wh(self, value):
        self.w, self.h = value

    # int
    @property
    def int_w(self):
        return math.floor(self.w)

    @property
    def int_h(self):
        return math.floor(self.h)

    @property
    def int_width(self):
        return self.int_w

    @property
    def int_height(self):
        return self.int_h

    @property
    def int_wh(self):
        return (self.int_width, self.int_height)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(w={self.w}, h={self.h})"



import pygame

from ..utils import Rect, Size


class Text:
    def __init__(self):
        self.font = None
        self.__color_background = None
        self.space_between_lines = 3
        self.__color_text = (0, 0, 0)
        self.text_data = []
        self.size = Size(0, 0)
        self.image = pygame.Surface(self.size.wh)

    def _set_color_text(self, color):
        self.__color_text = color

    def set_font_from_system(self, name, size, bold=False, italic=False):
        self.font = pygame.font.SysFont(name, size, bold, italic)

    def set_font_from_file(self, path_or_file_obj, size):
        self.font = pygame.font.Font(path_or_file_obj, size)

    def __process_line(self, line, index):
        to_return = []
        metrics = self.font.metrics(line)
        line_size = index * self.font.get_linesize()
        space_between_lines = index * self.space_between_lines
        for i, letter in enumerate(line):
            letter_surface = self.font.render(
                letter, True, self.__color_text, self.__color_background
            )
            rect = Rect.from_pygame_rect(letter_surface.get_rect())
            rect.y = line_size + space_between_lines
            if i > 0:
                rect.x += metrics[-1][4] + to_return[-1].rect.x
            else:
                rect.x = metrics[0][0]
            to_return.append(self.Letter(letter, letter_surface, rect))
        return to_return

    def _set_text(self, text):
        self.text_data.clear()
        lines = text.splitlines()
        for i, line in enumerate(lines):
            self.text_data.append(self.__process_line(line, i))
        self._draw()

    def _draw(self):
        w, h = 0, 0
        for i, line in enumerate(self.text_data):
            if len(line) > 0:
                w = max(w, line[-1].rect.right)
                if i + 1 == len(self.text_data):
                    for letter in line:
                        h = max(h, letter.rect.bottom)
        if self.image.get_size() != (w, h):
            self.image = pygame.Surface((w, h), pygame.SRCALPHA)
            self.size = Size(w, h)
        else:
            self.image.fill((0, 0, 0, 0))
        for line in self.text_data:
            for letter in line:
                self.image.blit(
                    letter.surface,
                    letter.rect.int_xy,
                )

    def _get_rendered_text(self, offset, size):
        right = offset.x + size.w
        bottom = offset.y + size.h
        if self.size.w < right:
            offset.x -= right - self.size.w
        if self.size.h < bottom:
            offset.y -= bottom - self.size.h

        if offset.x < 0:
            offset.x = 0
        if offset.y < 0:
            offset.y = 0

        if size.w > self.size.w:
            size.w = self.size.w
        if size.h > self.size.h:
            size.h = self.size.h

        return self.image.subsurface(pygame.Rect(offset.int_xy, size.int_wh))

    class Letter:
        def __init__(self, char, surface, rect) -> None:
            self.char = char
            self.surface = surface
            self.rect = rect



import pygame

from ..essentials import Scenes
from ..events import Events
from .window import Window
from .looptimer import LoopTimer


class Program:
    def __init__(self):
        pygame.init()

        self.window = Window()

        self.events = Events()
        self.scenes = Scenes()
        self.looptimer = LoopTimer()

    def run(self):
        while True:
            self.events.process()
            self.scenes.process()
            self.scenes.render()

            self.window.update()
            self.window.clear()

            self.looptimer.process()

    def exit(self):
        raise SystemExit(0)



from .event import Event


class System(Event):
    def __init__(self) -> None:
        super().__init__()
        self.__exit = False

    def request_exit(self):
        self.__exit = True

    def get_exit_requested(self):
        return self.__exit



import pygame

from ..utils import Position
from .inputevent import InputEvent


class Mouse(InputEvent):
    def __init__(self):
        super().__init__(mode="mouse")
        self.pos = Position()
        self.rel_pos = Position()
        self.wheel = Position()

    def init_process(self):
        self.wheel.xy = (0, 0)

    def process(self, event):
        match event.type:
            case pygame.MOUSEMOTION:
                self.pos.xy = event.pos
                self.rel_pos = event.rel

            case pygame.MOUSEBUTTONDOWN:
                self.pos.xy = event.pos
                self.pressed[event.button] = event
                self.keydown.append(event.button)

            case pygame.MOUSEBUTTONUP:
                self.pos.xy = event.pos
                del self.pressed[event.button]
                self.keyup.append(event.button)

            case pygame.MOUSEWHEEL:
                self.wheel.xy = event.x, event.y



import pygame


class LoopTimer:
    def __init__(self):
        self.deltatime = 0
        self.framerate = 0
        self.pygame_clock = pygame.time.Clock()

    def set_fps(self, fps):
        self.framerate = fps

    def process(self):
        self.deltatime = self.pygame_clock.tick(self.framerate) / 1000



from ..utils import Position, Size
from .image import Image
from .text import Text


class TextView(Image):
    def __init__(self, parent, pos=(0, 0), size=(0, 0)):
        super().__init__(parent, pos, size)
        self._create_image_from_rect()
        self.image.fill((0, 0, 0, 0))
        self.offset = Position(*pos)
        self.text_component = Text()
        self.scroll_speed = 50
        self.scrollable = False

    def process(self):
        super().process()
        wheel = self.program.events.mouse.wheel
        scroll = wheel.xy != (0, 0)
        if scroll:
            self.scroll(wheel.x * self.scroll_speed, wheel.y * self.scroll_speed)

    def set_text(self, text):
        self.text_component._set_text(text)
        self.image = self.text_component._get_rendered_text(
            self.offset, Size(*self.rect.wh)
        )

    def set_color_text(self, color):
        self.text_component._set_color_text(color)
        self.image = self.text_component._get_rendered_text(
            self.offset, Size(*self.rect.wh)
        )

    def scroll(self, x, y):
        if self.scrollable:
            self.offset.x -= x
            self.offset.y -= y

            self.image = self.text_component._get_rendered_text(
                self.offset, Size(*self.rect.wh)
            )



from copy import deepcopy
from os.path import abspath

import pygame
import pgk

from .image import Image


class RenderableImage(pgk.essentials.Node, Image):
    def __init__(self, parent, pos=(0, 0), size=(0, 0)):
        Image.__init__(self, pos, size)
        pgk.essentials.Node.__init__(self, parent)
        self.image_instance = None

    @property
    def image(self):
        return self.image_instance.image

    @image.setter
    def image(self, value):
        self.image_instance.image = value

    def process(self):
        pgk.essentials.Node.process(self)

    def render(self):
        self.image_instance.render(
            self.rect,
            self.program.window.image,
        )

    @classmethod
    def from_image(cls, parent, image_instance: Image, copy=False):
        return_img = cls(
            parent, image_instance.rect.xy, image_instance.image.get_size()
        )
        return_img.image_instance = image_instance
        if copy:
            return_img.image_instance = image_instance.__class__.from_image(
                image_instance
            )
        return_img.add_child(return_img.image_instance)
        return return_img

    @classmethod
    def from_image_cropping(cls, parent, image_instance, rect, pos=(0, 0), copy=False):
        return_img = cls(parent, pos, rect.wh)
        return_img.image_instance = Image.from_pygame_surface(
            image_instance.image.subsurface(rect.pygame_rect), pos, copy
        )
        return_img.add_child(return_img.image_instance)
        if copy:
            return_img.image_instance = deepcopy(return_img.image_instance)
        return return_img

    @classmethod
    def from_empty_image(cls, parent, rect):
        return_img = cls(parent, rect.xy, rect.wh)
        return_img.image_instance = Image.from_pygame_surface(
            pygame.Surface(rect.wh, pygame.SRCALPHA), rect.xy
        )
        return_img.add_child(return_img.image_instance)
        return return_img

    @classmethod
    def from_file(cls, parent, path, pos=(0, 0)):
        surface = pygame.image.load(abspath(path)).convert_alpha()
        return_img = cls(parent, pos, surface.get_size())
        return_img.image_instance = Image.from_pygame_surface(surface, pos)
        return_img.add_child(return_img.image_instance)
        return return_img



import pygame

from .keyboard import KeyBoard
from .mouse import Mouse
from .system import System


class Events:
    def __init__(self):
        self.keyboard = KeyBoard()
        self.mouse = Mouse()
        self.system = System()

    def process(self):
        self.keyboard.init_process()
        self.mouse.init_process()
        for e in pygame.event.get():
            match e.type:
                case pygame.QUIT:
                    self.system.request_exit()
                case pygame.KEYUP | pygame.KEYDOWN:
                    self.keyboard.process(e)
                case pygame.MOUSEMOTION | pygame.MOUSEBUTTONUP | pygame.MOUSEBUTTONDOWN | pygame.MOUSEWHEEL:
                    self.mouse.process(e)
                case _:
                    print(e)



import screeninfo

from .size import Size


class Monitor:
    monitors = screeninfo.get_monitors()

    @classmethod
    def get_primary_monitor_size(cls):
        for monitor in cls.monitors:
            if monitor.is_primary:
                return Size(monitor.width, monitor.height)



from os.path import abspath

import pygame

from ..essentials.basewidget import BaseWidget
from ..utils import Rect


class Image(BaseWidget):
    def __init__(self, pos=(0, 0), size=(0, 0)):
        self.rect = Rect(pos, size)

    def _create_image_from_rect(self):
        self.image = pygame.Surface(self.rect.wh, pygame.SRCALPHA)

    def load_file(self, path):
        self.image = pygame.image.load(abspath(path)).convert_alpha()
        self.rect.wh = self.image.get_size()

    def fill(self, color):
        self.image.fill(color)

    def render(self, rect_pos, image):
        super().render()
        image.image.blit(self.image, rect_pos.xy)

    @classmethod
    def from_pygame_surface(cls, surface, pos=(0, 0), copy_img=False):
        return_img = cls(pos, surface.get_size())
        return_img.image = surface
        if copy_img:
            return_img.image = return_img.image.copy()
        return return_img

    @classmethod
    def from_image(cls, image_instance, pos=(0, 0), copy_img=False):
        return_img = cls(pos, image_instance.image.get_size())
        return_img.image = image_instance.image
        if copy_img:
            return_img.image = return_img.image.copy()
        return return_img

    @classmethod
    def from_image_cropping(cls, image, rect, pos=(0, 0), copy_img=False):
        return_img = cls(pos, rect.wh)
        return_img.image = image.image.subsurface(rect.pygame_rect)
        if copy_img:
            return_img.image = return_img.image.copy()
        return return_img

    @classmethod
    def from_empty_image(cls, rect):
        return_img = cls(rect.xy, rect.wh)
        return_img._create_image_from_rect()
        return return_img

    @classmethod
    def from_file(cls, path, pos=(0, 0)):
        surface = pygame.image.load(abspath(path)).convert_alpha()
        return_img = cls(pos, surface.get_size())
        return_img.image = surface
        return return_img



import pygame

from .position import Position
from .size import Size


class Rect(Position, Size):
    def __init__(self, pos, size):
        Position.__init__(self, *pos)
        Size.__init__(self, *size)

    @classmethod
    def from_pygame_rect(cls, rect):
        return cls(rect.topleft, rect.size)

    def colliderect(self, rect):
        return (
            min(self.x, self.x + self.w) < max(rect.x, rect.x + rect.w)
            and min(self.y, self.y + self.h) < max(rect.y, rect.y + rect.h)
            and max(self.x, self.x + self.w) > min(rect.x, rect.x + rect.w)
            and max(self.y, self.y + self.h) > min(rect.y, rect.y + rect.h)
        )

    @property
    def pygame_rect(self):
        return pygame.Rect(self.int_xy, self.int_wh)

    # basic
    @property
    def top(self):
        return self.y

    @property
    def bottom(self):
        return self.y + self.h

    @property
    def left(self):
        return self.x

    @property
    def right(self):
        return self.x + self.w

    @top.setter
    def top(self, value):
        self.y = value

    @bottom.setter
    def bottom(self, value):
        self.y = value - self.h

    @left.setter
    def left(self, value):
        self.x = value

    @right.setter
    def right(self, value):
        self.x = value - self.w

    # advanced
    @property
    def topleft(self):
        return self.left, self.top

    @property
    def topright(self):
        return self.right, self.top

    @property
    def bottomleft(self):
        return self.left, self.bottom

    @property
    def bottomright(self):
        return self.right, self.bottom

    def __repr__(self):
        class_name = self.__class__.__name__
        x, y = self.xy
        w, h = self.wh
        return f"<{class_name}(x={x}, y={y}, width={w}, height={h})>"



from os.path import abspath

from ..essentials.basewidget import BaseWidget
from .image import Image


class Sprite(BaseWidget):
    def __init__(self):
        self.source = self.Source()

    def add_image(self, path_or_image, reference):
        type_image_arg = type(path_or_image)
        if type_image_arg == str:
            self.source.add_image(Image.from_file(abspath(path_or_image)), reference)
        elif type_image_arg == Image:
            image = path_or_image
            self.source.add_image(image, reference)

    def add_spritesheet(self, path_or_image, reference):
        type_image_arg = type(path_or_image)
        if type_image_arg == str:
            self.source.add_spritesheet(
                Image.from_file(abspath(path_or_image)), reference
            )
        elif type_image_arg == Image:
            image = path_or_image
            self.source.add_spritesheet(image, reference)

    def add_from_sprite_sheet(
        self, rect, spritesheet_name, new_image_reference=None, pos=(0, 0), copy=False
    ):
        new_image = Image.from_image_cropping(
            self.source.get_spritesheet(spritesheet_name), rect, pos, copy
        )
        self.source.add_image_from_spritesheet(
            new_image, spritesheet_name, new_image_reference
        )

    class Source:
        def __init__(self):
            self.muted = False
            self.images = {}
            self.spritesheets = {}
            self.__image_unnamed = 0
            self.__spritesheet_unnamed = 0

        def add_image(self, image, reference=None):
            if reference is None:
                while (
                    name_temp := f"image{self.__image_unnamed}"
                ) in self.images.keys():
                    self.__image_unnamed += 1
                reference = name_temp
            self.images[reference] = image
            if not self.muted:
                print(f"Image '{reference}' loaded successfully")

        def add_spritesheet(self, image, reference=None):
            if reference is None:
                while (
                    name_temp := f"spritesheet{self.__spritesheet_unnamed}"
                ) in self.images.keys():
                    self.__spritesheet_unnamed += 1
                reference = name_temp
            self.spritesheets[reference] = image
            if not self.muted:
                print(f"Spritesheet '{reference}' loaded successfully")

        def add_image_from_spritesheet(self, image, spritesheet_name, reference):
            if reference is None:
                i = 0
                while (name_temp := f"{spritesheet_name}.{i}") in self.images.keys():
                    self.self.i += 1
                reference = name_temp
            self.images[reference] = image
            if not self.muted:
                print(
                    f"Image '{reference}' loaded successfully from '{spritesheet_name}' spritesheet"
                )

        def get_image(self, reference):
            return self.images[reference]

        def get_spritesheet(self, reference):
            return self.spritesheets[reference]



import math

import pygame

from ..utils import Position
from .image import Image


class FilledImage(Image):
    def __init__(self, to_fill, pos=(0, 0), size=(0, 0)):
        super().__init__(pos, size)
        self.image = pygame.Surface(size, pygame.SRCALPHA)
        self.filled_surface = None
        self.offset = Position()
        self.to_fill = to_fill

        self.pre_render()

    def render(self, rect_pos, image):
        self.fill((0, 0, 0, 0))
        self.image.blit(self.filled_surface, self.offset.xy)
        super().render(rect_pos, image)

    def pre_render(self):
        self.fill((0, 0, 0, 0))
        lines = math.ceil(self.rect.h / self.to_fill.rect.h) + 1
        columns = math.ceil(self.rect.w / self.to_fill.rect.w) + 1

        self.filled_surface = pygame.Surface(
            (columns * self.to_fill.rect.w, lines * self.to_fill.rect.h),
            pygame.SRCALPHA,
        )

        for y in range(lines):
            y = y * self.to_fill.rect.h
            for x in range(columns):
                x = x * self.to_fill.rect.w
                self.filled_surface.blit(self.to_fill.image, (x, y))



from ..essentials import Node
from ..utils import Position
from .textview import TextView


class EditText(TextView):
    def __init__(self, parent, pos=(0, 0), size=(0, 0)):
        super().__init__(parent, pos, size)
        self.cursor = self.Cursor(self)
        self.add_child(self.cursor)

    class Cursor(Node):
        def __init__(self, parent):
            super().__init__(parent)
            self.pos = Position()



from .edittext import EditText
from .filledimage import FilledImage
from .image import Image
from .renderableimage import RenderableImage
from .sprite import Sprite
from .text import Text
from .textview import TextView



import math


class Position:
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    @property
    def xy(self):
        return (self.x, self.y)

    @xy.setter
    def xy(self, value):
        self.x, self.y = value

    # int
    @property
    def int_x(self):
        return math.floor(self.x)

    @property
    def int_y(self):
        return math.floor(self.y)

    @property
    def int_xy(self):
        return (self.int_x, self.int_y)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(x={self.x}, y={self.y})"



import pygame

from ..nodes import Image
from ..utils import Size


class Window:
    def __init__(self):
        self.image = None

    def set_size(self, width, height):
        self.image = Image.from_pygame_surface(
            pygame.display.set_mode((width, height), pygame.DOUBLEBUF)
        )

    def set_title(self, title):
        pygame.display.set_caption(title)

    def landscape(self):
        return self.image.rect.w > self.image.rect.h

    def update(self):
        pygame.display.update()

    def clear(self):
        self.image.fill((0, 0, 0))

    @property
    def size(self):
        return Size(*self.image.rect.wh)



import pygame

from .event import Event


class InputEvent(Event):
    MOUSE_KEYS = {
        "left": pygame.BUTTON_LEFT,
        "middle": pygame.BUTTON_MIDDLE,
        "right": pygame.BUTTON_RIGHT,
    }

    def __init__(self, mode):
        super().__init__()
        self.pressed = {}
        self.keydown = []
        self.keyup = []
        self.mode = mode

    def init_process(self):
        self.keyup.clear()
        self.keydown.clear()

    def get_key_code(self, name):
        match self.mode:
            case "keyboard":
                try:
                    return pygame.key.key_code(name.lower())
                except ValueError:
                    raise ValueError(f"Key {name} not found")
            case "mouse":
                try:
                    return self.MOUSE_KEYS[name]
                except KeyError:
                    raise ValueError(f"Key {name} not found")

    def get_key_down(self, name):
        cod = self.get_key_code(name)
        return cod in self.keydown

    def get_key_up(self, name):
        cod = self.get_key_code(name)
        return cod in self.keyup

    def get_pressed(self, name):
        cod = self.get_key_code(name)
        return cod in self.pressed or self.get_key_down(name)



class Scenes:
    def __init__(self):
        self.current_scene = None
        self.scenes = {}

    def process(self):
        self.current_scene.process()

    def render(self):
        self.current_scene.render()

    def set_current_scene(self, scene, scene_name=None):
        self.current_scene = scene
        if scene_name is None:
            scene_name = scene.__class__.__name__
        self.scenes[scene_name] = scene.__class__



from setuptools import setup

setup(
    name="pgk",
    version="0.1.3",
    license="MIT",
    author="ExtraMobs",
    author_email="extramobs@hotmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.10",
    ],
    url="https://github.com/ExtraMobs/pgk",
    keywords="PyGameKit",
    python_requires=">=3.10",
    install_requires=[
        "pygame",
        "screeninfo"
    ]
)



import pygame

from .inputevent import InputEvent


class KeyBoard(InputEvent):
    def __init__(self):
        super().__init__(mode="keyboard")

    def process(self, event):
        match event.type:
            case pygame.KEYDOWN:
                self.pressed[event.key] = event
                self.keydown.append(event.key)
            case pygame.KEYUP:
                del self.pressed[event.key]
                self.keyup.append(event.key)

    def get_key_code(self, name):
        return super().get_key_code(name)



from .basewidget import BaseWidget


class Widget(BaseWidget):
    def __init__(self, program):
        self.program = program
        self.__children = []

    def process(self):
        for child in self.__children:
            child.process()

    def render(self):
        for child in self.__children:
            child.render()

    def add_child(self, *widget):
        self.__children.extend(widget)

    def set_child(self, index, new_widget):
        self.__children[index] = new_widget

    def remove_child(self, widget):
        self.__children.remove(widget)

    def get_child(self, index):
        return self.__children[index]

    @property
    def children(self):
        return tuple(self.__children)
